# -*- coding: utf-8 -*-

from __future__ import with_statement

import csv
import inspect
import operator
import os.path
import re
import smtplib
import subprocess
import sys
import threading
import time
import traceback
import types

from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Utils import COMMASPACE, formatdate
from email import Encoders

class MailSender:
    '''Envia un mail con archivos attached.'''
    def __init__(self, server, remitente, auth=None, expl_from=None):
        self.server = server
        self.remitente = remitente
        self.auth = auth
        self.expl_from = expl_from

    def envia(self, to, subject, text, files=[]):
        assert type(files)==list
        if not isinstance(to, list):
            to = [to]

        msg = MIMEMultipart()
        fromt = self.expl_from if self.expl_from is not None else self.remitente
        msg['From'] = fromt
        msg['To'] = COMMASPACE.join(to)
        msg['Date'] = formatdate(localtime=True)
        msg['Subject'] = subject

        msg.attach( MIMEText(text, _charset="utf8") )

        for file in files:
            part = MIMEBase('application', "octet-stream")
            part.set_payload( open(file,"rb").read() )
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(file))
            msg.attach(part)

        smtp = smtplib.SMTP(self.server)
        if self.auth is not None:
            auth_user, auth_pass = self.auth
            smtp.ehlo()
            smtp.starttls()
#            smtp.ehlo()
            smtp.login(auth_user, auth_pass)
        smtp.sendmail(self.remitente, to, msg.as_string())
        smtp.close()
        return


class Avance:
    '''Muestra un avance amigable al stdout.

    Para usarlo, instanciar la clase con el total, y luego, por
    cada iteracion llamar a step() con lo recorrido (1 por default)
    '''
    def __init__(self, total):
        self.total = total / 100.0
        self.antval = None
        self.cant = 0
        self.tant = time.time()

    def step(self, paso=1):
        self.cant += paso
        pt = int(self.cant / self.total)
        if pt == self.antval:
            return
        self.antval = pt
        tnow = time.time()
        sys.stdout.write("\r    %3d %%   %d segs" % (pt, tnow-self.tant))
        if pt >= 100:
            sys.stdout.write("\n")
        sys.stdout.flush()
        self.tant = tnow

def ejecutar(cmd, separado=True, shell=False):
    '''Ejecuta algo y devuelve el stdout de lo ejecutado.

    Ej: ejecutar("ps -eaf")
    '''
    p = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, shell=shell)
    t = p.stdout.read()
    p.stdout.close()

    if separado:
        t = t.split("\n")
    return t


def ejecuta_mostrando(cmd):
    '''Ejecuta algo y va mostrando la salida mientras tanto.

    Ej: ejecutar("ps -eaf")
    '''
    p = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    res = []
    while True:
        t = p.stdout.readline()
        if not t:
            break
        sys.stdout.write(t)
        res.append(t)
    p.stdout.close()
    return res


def fechaAnterior(formato, atras=24*3600):
    '''Devuelve la fecha con algunos segundos restados, en el formato
    especificado. Por default, devuelve ayer.'''
    ayer = time.localtime(time.time()-atras)
    return time.strftime(formato, ayer)


def muestraCantDict(d, limite=None, prefijo=''):
    '''Muestra un diccionario donde los valores son cantidades.

    Ordena de mayor a menor, y muestra encolumnado en función del largo del valor.
    '''

    maxl = max(len(str(x)) for x in d.itervalues())
    data = sorted(d.iteritems(), key=operator.itemgetter(1), reverse=True)
    for i, (k,v) in enumerate(data):
        if limite is not None and i > limite:
            break
        print "%s%*d  %s" % (prefijo, maxl, v, k)

class Indent(object):
    indents = {}

    def __init__(self, multip='..'):
        self.multip = multip

    def up(self):
        th = threading.currentThread().getName()
        self.indents[th] = self.indents.get(th, 0) + 1

    def down(self):
        th = threading.currentThread().getName()
        self.indents[th] = self.indents.get(th, 0) - 1

    def __str__(self):
        th = threading.currentThread().getName()
        return self.multip * self.indents.get(th, 0)

def logger(filename, muestra_modulo=False):
    '''Returns a decorator that logs.

    It logs the call (with the args, and from where it was called), and the
    returned value or the raised exception.  In any case, it also puts the
    date, time, and the execution thread.  It always it's a line per record.

    The called object is identified with the function name if it's a function,
    or the class and method name if it's a method.  Optionally, it can be
    prefixed with the module where it's defined.

    @param filename: the name of the file where log to.
    @param muestra_modulo: if the module name will be added in the logs

    Usage example:

      l = logger("/tmp/prueba.log")
      @l
      def f(...):
          ...
    '''
    indent = Indent()

    class Logger(object):

        def __init__(self, func):
            self.orig_func = func

            if isinstance(func, types.UnboundMethodType):
                fcls = func.im_class
                modulo = fcls.__module__
                self.fname = "%s.%s" % (fcls.__name__, func.func_name)
            else:
                modulo = func.__module__
                self.fname = func.func_name
            if muestra_modulo:
                self.fname = "%s.%s" % (modulo, self.fname)

        def __get__(self, obj, type=None):
            return self.__class__(self.orig_func.__get__(obj, type))

        def logthis(self, text):
            t = time.time()
            fechahora = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
            msegs = ("%.3f" % divmod(t, 1)[1])[2:]
            th = threading.currentThread().getName()
            with open(filename, "a") as fh:
                fh.write("%s.%s (%s): %s%s\n" % (
                                    fechahora, msegs, th, indent, text))

        def __call__(self, *args, **kwargs):
            cframe = inspect.stack()[1]
            self.logthis("%r called with %s %s (from %r @ %s # %d)" % (
                self.fname, args, kwargs, cframe[3], cframe[1], cframe[2]))
            indent.up()

            try:
                r = self.orig_func(*args, **kwargs)
            except Exception, err:
                tb = traceback.extract_tb(sys.exc_info()[2])[-1]
                errtype = err.__class__.__name__
                message = err.message or err.args
                indent.down()
                self.logthis("%r raised %s (# %d): %s" % (
                                    self.fname, errtype, tb[1], message))
                raise

            indent.down()
            self.logthis("%r returned %r" % (self.fname, r))
            return r

    return Logger


def fecha2desde(lodijo, ahora=None, allow_future=False):
    """Devuelve el tiempo que pasó como un string simpático.

    >>> import datetime
    >>> lodijo = datetime.datetime(2009, 07, 26, 13, 35).timetuple()
    >>> fecha2desde(lodijo, lodijo)
    u'menos de un minuto'
    >>> ahora = datetime.datetime(2009, 07, 26, 13, 35, 20).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'menos de un minuto'
    >>> ahora = datetime.datetime(2009, 07, 26, 13, 30).timetuple()
    >>> fecha2desde(lodijo, ahora)
    Traceback (most recent call last):
    ...
    ValueError: Lo dijo en el futuro! lodijo=1248626100.0, ahora=1248625800.0
    >>> ahora = datetime.datetime(2009, 07, 26, 13, 36).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'un minuto'
    >>> ahora = datetime.datetime(2009, 07, 26, 13, 37).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'2 minutos'
    >>> ahora = datetime.datetime(2009, 07, 26, 13, 38).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'3 minutos'
    >>> ahora = datetime.datetime(2009, 07, 26, 14, 34).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'59 minutos'
    >>> ahora = datetime.datetime(2009, 07, 26, 14, 35).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'una hora'
    >>> ahora = datetime.datetime(2009, 07, 26, 14, 36).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'una hora'
    >>> ahora = datetime.datetime(2009, 07, 26, 15, 10).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'2 horas'
    >>> ahora = datetime.datetime(2009, 07, 26, 16, 35).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'3 horas'
    >>> ahora = datetime.datetime(2009, 07, 27, 13, 35).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'un d\\xeda'
    >>> ahora = datetime.datetime(2009, 07, 27, 14, 35).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'un d\\xeda'
    >>> ahora = datetime.datetime(2009, 07, 28, 3, 30).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'2 d\\xedas'
    >>> ahora = datetime.datetime(2009, 07, 30, 3, 30).timetuple()
    >>> fecha2desde(lodijo, ahora)
    u'4 d\\xedas'
    """

    # convertimos a segundos
    lodijo = time.mktime(lodijo)
    if ahora is None:
        ahora = time.mktime(time.gmtime())
    else:
        ahora = time.mktime(ahora)

    # controlamos futuro
    if lodijo > ahora:
        if allow_future:
            return None
        m = "Lo dijo en el futuro! lodijo=%s, ahora=%s" % (lodijo, ahora)
        raise ValueError(m)

    diff = ahora - lodijo
    if diff < 60:
        return u"menos de un minuto"

    # pasamos a manejar por minuto!
    lodijo = lodijo / 60
    ahora = ahora / 60
    diff = int(round(ahora - lodijo))
    if diff in (0, 1) :
        return u"un minuto"
    if diff < 60:
        return u"%s minutos" % (diff,)

    # pasamos a manejar por hora!
    lodijo = lodijo / 60
    ahora = ahora / 60
    diff = int(round(ahora - lodijo))
    if diff in (0, 1) :
        return u"una hora"
    if diff < 24:
        return u"%s horas" % (diff,)

    # pasamos a manejar por día!
    lodijo = lodijo / 24
    ahora = ahora / 24

    diff = int(round(ahora - lodijo))
    if diff in (0, 1) :
        return u"un día"
    return u"%s días" % (diff,)


def realtime():
    """Devuelve la hora con milisegundos."""
    t = time.time()
    full = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
    mseg = repr(t % 1)[1:5]
    return full + mseg


def open_pg_output(fh):
    """Return the pg lines as a dict."""
    for lin in fh:
        if sorted(set(lin.strip())) in (['-'], ['+', '-']):
            break
        header = lin
    else:
        raise ValueError("Header not found.")

    header = [x.strip() for x in header.split('|')]
    if len(header) != len(set(header)):
        s = set()
        fixed = []
        c = 2
        for h in header:
            if h in s:
                h += str(c)
                c += 1
            s.add(h)
            fixed.append(h)
        header = fixed

    reader = csv.reader(fh, delimiter="|")
    for lin in reader:
        content = [x.strip() for x in lin]
        if len(content) != len(header):
            continue
        yield dict(zip(header, content))


def open_pg_multiple(fh):
    """Return the pg lines as a dict."""

    def _get_header():
        for lin in fh:
            if sorted(set(lin.strip())) in (['-'], ['+', '-']):
                break
            header = lin
        else:
            return None

        header = [x.strip() for x in header.split('|')]
        if len(header) != len(set(header)):
            s = set()
            fixed = []
            c = 2
            for h in header:
                if h in s:
                    h += str(c)
                    c += 1
                s.add(h)
                fixed.append(h)
            header = fixed
        return header

    def _retrieve_data(header):
        reader = csv.reader(fh, delimiter="|")
        for lin in reader:
            content = [x.strip() for x in lin]
            if not content:
                break
            yield dict(zip(header, content))

    results = []
    while True:
        # sync to first header
        header = _get_header()
        if header is None:
            break

        result = list(_retrieve_data(header))
        last_item = result[-1].values()[0]
        result = result[:-1]
        m = re.match("\((\d+) rows\)", last_item)
        if not m:
            raise ValueError("Bad last item: " + repr(last_item))
        qvals = int(m.groups()[0])
        if len(result) != qvals:
            raise ValueError("Bad result lenght: should=%d got=%d" % (
                             qvals, len(result)))

        results.append(result)

    return results


if __name__ == "__main__":
    import doctest
    doctest.testmod()


